import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/*
    Main class for Retrieve. Calls RetrieveSimple (k-NN retrieval), retrieves a list of armies, then does complex matching to retrieve the best army from the k armies listed.
 */
public class Retrieve {

    //These two variables determine point size and faction, and size of k (How many armies the simple retrieval retrieves).
    public static String wanted = "None";
    public static double points = 1850.0;
    public static int k = 6;
    //Terrain is in order: Building %, Difficult Terrain %, Dangerous Terrain %, Impassable Terrain %.
    public static List<Double> terrain = Arrays.asList(15.0, 10.0, 0.0, 0.0);

    //Main method
    public static Army RetrieveBest(ArrayList<Army> army, Army problemArmy){

        double similarity = 0;
        int bestArmy = -1;

        //These variables are purely used for explanations
        double bestARating = 0;
        double bestSRating = 0;
        double bestMoveRating = 0;
        double bestShootRating = 0;
        double bestPsyRating = 0;
        double bestAssaultRating = 0;
        double bestFavourRating = 0;

        //Simple retrieval is performed, with k, wanted, and points variables assigned at the start of this class.
        long retrieveTime = System.currentTimeMillis();
        ArrayList<Army> best = RetrieveSimple.Retrieve(army, k, wanted, points);
        if (!CBRMain.retrievalTimeTrue) {
            CBRMain.retrieveTime = System.currentTimeMillis() - retrieveTime;
            CBRMain.retrievalTimeTrue = true;
        }

        double[] calculated = armyCalculation(problemArmy);
        double pavgToughness = calculated[0];
        double pavgWToughness = calculated[1];
        double pavgAS = calculated[2];
        double pavgWS = calculated[3];

        //Complex retrieval is now performed, checking in depth each army vs. the problem army.
        for (int i = 0; i < best.size(); i++){

            Army a = best.get(i);

            /*
                Calculate Ratio of Army Rating, this is an indication (though a fairly weak one) of how armies match up.
                New armies have 1500 rating, so they may be very strong or very weak. However, given time this should play a small part in similarity
             */
            double ratioOfRating = ratioOfRating(a, problemArmy);

            /*
                Calculate the Ratio of Squad Ratings. Much like the army ratings, it is an indication of how individual squads as a whole match up.
             */
            double ratioOfSquads = ratioOfSquads(a, problemArmy);

            /*
                Calculate Ratio of Army Strengths.
                This takes the ratio of enemy units, and compares them to the units from the army and their equipment.
             */

            double[] calculated2 = armyCalculation(a);
            double savgToughness = calculated2[0];
            double savgWToughness = calculated2[1];
            double savgAS = calculated2[2];
            double savgWS = calculated2[3];

            double ratioOfMovement = movementRate(a, problemArmy);
            double ratioOfPsychic = psychicRate(a, problemArmy);
            double ratioOfShooting = shootingRate(a, problemArmy, pavgToughness, savgToughness, pavgAS, savgAS);
            double ratioOfAssault = assaultRate(a, problemArmy, pavgWToughness, savgWToughness, pavgAS, savgAS, pavgWS, savgWS);

            double ratioOfStrength = (ratioOfMovement*CBRMain.movementWeight + ratioOfPsychic*CBRMain.psychicWeight + ratioOfShooting*CBRMain.shootingWeight + ratioOfAssault*CBRMain.assaultWeight)
                                     / (CBRMain.movementWeight+CBRMain.psychicWeight+CBRMain.shootingWeight+CBRMain.assaultWeight);

            /*
                Ratio of Favor.
                This is the ratio that favors battle forged armies, and also checks for a warlord with a warlord trait.
                Other rules can be added later, if time permits.
             */
            double ratioOfFavor = detachmentCheck(a, problemArmy);

            /*
                Similarity should be equal:
                Ratio of Army Ratings (Weight 1.0) + Ratio of Army Strengths (Weight 5.0) + Ratio of Factions (1.0) + Ratio of Favor (3.0)
                Divided by 10, to accommodate the weights.
             */
            double temp = (ratioOfRating * 1.0 + ratioOfSquads*1.0 + ratioOfStrength * 6.0 + ratioOfFavor * 3.0)/11.0;

            /*
                Set the highest similarity and the best army. This is the army we will return.
             */

            if (temp > similarity){
                similarity = temp;
                bestArmy = i;
                bestARating = ratioOfRating;
                bestSRating = ratioOfSquads;
                bestMoveRating = ratioOfMovement;
                bestShootRating = ratioOfShooting;
                bestPsyRating = ratioOfPsychic;
                bestAssaultRating = ratioOfAssault;
                bestFavourRating = ratioOfFavor;
            }

        }

        CBRMain.retrievalExplanations.add("The best army retrieved has a " + String.format("%.4g", similarity*100.0) + "% chance to win this match.");
        explanations(best.get(bestArmy), problemArmy, similarity, bestARating, bestSRating, bestMoveRating, bestShootRating, bestPsyRating, bestAssaultRating, bestFavourRating);
        if (CBRMain.winChanceBeforeReuse == 0){
            CBRMain.winChanceBeforeReuse = similarity;
        }
        else if (CBRMain.winChanceAfterReuse == 0){
            CBRMain.winChanceAfterReuse = similarity;
        }
        else{
            CBRMain.winChanceAfterRevise = similarity;
        }

        return best.get(bestArmy);
    }

    /*
        This class calculates the movement capability of an army. An army that is more mobile has an advantage.
        Both the standard movement and the maximum movement is considered, and the ratio is returned.
     */
    public static double movementRate(Army a, Army p){

        int unitCount = 0;
        int transportCount = 0;
        int infantryCount = 0;
        double normalMovement = 0;
        double expeditousMovement = 0;
        double similarity = 0;

        double finalNormal;
        double finalExpeditous;

        /*
           Uses general domain knowledge to check for movement speeds, but also checks for transports and infantry.
           It is assumed that transports are always used to the best of ones ability (that is to say, to transport infantry).
           The type of transport is (currently) ignored.
         */
        for (int i = 0; i < p.getSquad().size(); i++){
            for (int j = 0; j < p.getSquad().get(i).getUnit().size(); j++){
                if (p.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    normalMovement += GeneralKnowledge.movementKnowledge(p.getSquad().get(i).getUnit().get(j).getType(), true);
                    expeditousMovement += GeneralKnowledge.movementKnowledge(p.getSquad().get(i).getUnit().get(j).getType(), false);
                    unitCount++;
                    if (p.getSquad().get(i).getUnit().get(j).getType().contains("Transport")) {
                        transportCount++;
                    }
                    if (p.getSquad().get(i).getUnit().get(0).getType().contains("Infantry")) {
                        infantryCount++;
                    }
                }
            }
        }

        expeditousMovement += normalMovement;

        if ((infantryCount > transportCount) && (transportCount != 0)){
            normalMovement = normalMovement + infantryCount*3;
        }
        else{
            normalMovement = normalMovement + transportCount*3;
        }

        normalMovement = normalMovement/unitCount;
        expeditousMovement = expeditousMovement/unitCount;

        finalNormal = normalMovement*terrainRate(p, terrain);
        finalExpeditous = expeditousMovement*terrainRate(p, terrain);

        //Do the same for solution army
        normalMovement = 0;
        expeditousMovement = 0;
        unitCount = 0;
        infantryCount = 0;
        transportCount = 0;

        for (int i = 0; i < a.getSquad().size(); i++){
            for (int j = 0; j < a.getSquad().get(i).getUnit().size(); j++) {
                if (a.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    normalMovement += GeneralKnowledge.movementKnowledge(a.getSquad().get(i).getUnit().get(j).getType(), true);
                    expeditousMovement += GeneralKnowledge.movementKnowledge(a.getSquad().get(i).getUnit().get(j).getType(), false);
                    unitCount++;
                    if (a.getSquad().get(i).getUnit().get(j).getType().contains("Transport")) {
                        transportCount++;
                    }
                    if (a.getSquad().get(i).getUnit().get(0).getType().contains("Infantry")) {
                        infantryCount++;
                    }
                }
            }
        }

        expeditousMovement += normalMovement;

        if ((infantryCount > transportCount) && (transportCount != 0)){
            normalMovement = normalMovement + infantryCount*3;
        }
        else{
            normalMovement = normalMovement + transportCount*3;
        }

        normalMovement = normalMovement/unitCount;
        expeditousMovement = expeditousMovement/unitCount;

        finalNormal -= normalMovement*terrainRate(p, terrain);
        finalExpeditous -= expeditousMovement*terrainRate(p, terrain);

        /*
            Currently, normal movement is weighted at 70% while expeditious is weighted at 30%.
            If an entire army on average can move 1 inch move, it is much more maneuverable than the opposing army.
         */
        if (finalNormal < -3){
            similarity += 0.9;
        }

        if (finalExpeditous < -3){
            similarity += 0.1;
        }

        if (finalNormal < 3 && finalNormal > -3){
            similarity += 0.9*(0.5+0.166*(-finalNormal));
        }

        if (finalExpeditous < 3 && finalExpeditous > -3){
            similarity += 0.1*(0.5+0.166*(-finalExpeditous));
        }

        return similarity;
    }

    public static double psychicRate(Army a, Army p){

        double similarity = 0;
        int problemPsychic = 0;
        int solutionPsychic = 0;

        /*
           Counts the amount of Psychic mastery levels each army has and compares them.
           The psychic power ratio is just a measure of psychic strength, it is up to the players to use it correctly.
         */

        for (int i = 0; i < p.getSquad().size(); i++){
            for (int j = 0; j < p.getSquad().get(i).getUnit().size(); j++){
                if (p.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    for (int k = 0; k < p.getSquad().get(i).getUnit().get(j).getRules().size(); k++) {
                        if (p.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Psyker") ||
                                p.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Psychic Pilot") ||
                                p.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Brotherhood of Psykers") ||
                                p.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Brotherhood of Sorcerers")) {
                            problemPsychic += Integer.parseInt(p.getSquad().get(i).getUnit().get(j).getRules().get(k).substring(p.getSquad().get(i).getUnit().get(j).getRules().get(k).length() - 1, p.getSquad().get(i).getUnit().get(j).getRules().get(k).length()));
                        }
                    }
                }
            }
        }

        for (int i = 0; i < a.getSquad().size(); i++){
            for (int j = 0; j < a.getSquad().get(i).getUnit().size(); j++){
                if (a.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    for (int k = 0; k < a.getSquad().get(i).getUnit().get(j).getRules().size(); k++) {
                        if (a.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Psyker") ||
                                a.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Psychic Pilot") ||
                                a.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Brotherhood of Psykers") ||
                                a.getSquad().get(i).getUnit().get(j).getRules().get(k).contains("Brotherhood of Sorcerers")) {
                            solutionPsychic += Integer.parseInt(a.getSquad().get(i).getUnit().get(j).getRules().get(k).substring(a.getSquad().get(i).getUnit().get(j).getRules().get(k).length() - 1, a.getSquad().get(i).getUnit().get(j).getRules().get(k).length()));
                        }
                    }
                }
            }
        }

        //in the case of no psychic abilities the armies are tied, and similarity is 0.5.
        if (solutionPsychic == 0 && problemPsychic == 0){
            similarity = 0.5;
        }
        //If the difference is 3 or more, an army has a major advantage
        else if (solutionPsychic - problemPsychic >= 3){
            similarity = 1.0;
        }
        else if (solutionPsychic - problemPsychic > -3 && solutionPsychic - problemPsychic < 3){
            similarity = ((solutionPsychic - problemPsychic)*(0.16666))+0.5;
        }

        return similarity;
    }

    public static double shootingRate(Army a, Army p, double proavgToughness, double solavgToughness, double proavgAS, double solavgAS) {

        /*
            The shooting phase is one of the most complex rates do determine.
            The effectiveness rating is determined by Strength vs. Toughness, AP vs. Average AS, S vs. Armor is counted as toughness for the purposes of this test.
            This is modified by range: 24 - 36 * 1.0, 12-24 * 0.5, 0-12 * 0.25, 36+ * 2.0
            Then the best effectiveness is chosen for a unit, until all weapons have been used. This is then added together as a global effectiveness rating. It's ratio is the strength of the phase.
         */

        double problemEfficiency;
        double solutionEfficiency;
        double proBarrage;
        double solBarrage;

        //Then we determine the approximate effectiveness for problem army.
        double[] res = shootingCalculation(p, solavgToughness, solavgAS);
        problemEfficiency = res[0];
        proBarrage = res[1];
        //Then we determine the approximate effectiveness for solution army.
        double[] res2 = shootingCalculation(a, proavgToughness, proavgAS);
        solutionEfficiency = res2[0];
        solBarrage = res2[1];

        //We calculate how much cover influences our shooting (The more cover, the less our overall effectiveness)
        problemEfficiency = problemEfficiency - proBarrage;
        problemEfficiency*=shootingWeight(terrain);
        problemEfficiency+=proBarrage;

        solutionEfficiency = solutionEfficiency - solBarrage;
        solutionEfficiency*=shootingWeight(terrain);
        solutionEfficiency+=solBarrage;

        //Finally we return the value for the problem efficiency vs. solution efficiency.
        if (problemEfficiency == solutionEfficiency){
            return 0.5;
        }
        else if (problemEfficiency*2 < solutionEfficiency){
            return 1.0;
        }
        else if (problemEfficiency/2 > solutionEfficiency){
            return 0.0;
        }
        else{
            return 0.5 * (solutionEfficiency/problemEfficiency);
        }
    }

    public static double assaultRate(Army a, Army p, double proavgToughness, double solavgToughness, double proavgAS, double solavgAS, double proavgWS, double solavgWS) {

        /*
            The assault phase is one of the most complex rates do determine.
            We assume that an army will assault effectively (that is, an army will make their charge rolls) 100% of the time.
            The effectiveness rating is determined by WS vs. WS, Strength vs. Toughness, AP vs. Average AS, Initiative vs. Initiative.
            Then the best effectiveness is chosen for a unit, until all weapons have been used. This is then added together as a global effectiveness rating. It's ratio is the strength of the phase.
         */

        double problemEfficiency;
        double solutionEfficiency;

        //Then we determine the approximate effectiveness for problem army.
        problemEfficiency = assaultCalculation(p, solavgWS, solavgToughness, solavgAS);
        //Then we determine the approximate effectiveness for solution army.
        solutionEfficiency = assaultCalculation(a, proavgWS, proavgToughness, proavgAS);

        //Finally we return the value for the problem efficiency vs. solution efficiency.
        problemEfficiency*=assaultWeight(terrain);
        solutionEfficiency*=assaultWeight(terrain);

        if (problemEfficiency == solutionEfficiency){
            return 0.5;
        }
        else if (problemEfficiency*2 < solutionEfficiency){
            return 1.0;
        }
        else if (problemEfficiency/2 > solutionEfficiency){
            return 0.0;
        }
        else{
            return 0.5 * (solutionEfficiency/problemEfficiency);
        }
    }

    /*
        Units in detachements gain special abilities, that will help them perform better.
        A core detachement is always assumed, which consists of 1 HQ and 2 Assault Units.
        The rest of the detachements are presented in the General Knowledge and are filled up first using the rest of the units.
        If one army has a warlord and the other does not, the other army is penalized.
        If one army has detachements and the other does not, we measure the amount of units outside of detachements. The ratio becomes the remainder of the similarity test.
     */
    public static double detachmentCheck(Army a, Army p){

        String problemfaction = p.getSquad().get(0).getUnit().get(0).getFaction();
        String solutionfaction = p.getSquad().get(0).getUnit().get(0).getFaction();

        double problemRatio = p.getSquad().size();
        double solutionRatio = a.getSquad().size();

        problemRatio = GeneralKnowledge.detachments(problemfaction, p) / problemRatio;
        solutionRatio = GeneralKnowledge.detachments(solutionfaction, a) / solutionRatio;

        boolean warlord = false;
        for (int i = 0; i < p.getSquad().size(); i++){
            for (int j = 0; j < p.getSquad().get(i).getUnit().size(); j++) {
                if (p.getSquad().get(i).getUnit().get(j).getCount() > 0 && p.getSquad().get(i).getUnit().get(j).getType().contains("Character")) {
                    warlord = true;
                }
            }
        }

        if (!warlord){
            problemRatio = Math.min(problemRatio-0.1, 0);
        }

        warlord = false;
        for (int i = 0; i < a.getSquad().size(); i++){
            for (int j = 0; j < a.getSquad().get(i).getUnit().size(); j++) {
                if (a.getSquad().get(i).getUnit().get(j).getCount() > 0 && a.getSquad().get(i).getUnit().get(j).getType().contains("Character")) {
                    warlord = true;
                }
            }
        }

        if (!warlord){
            solutionRatio = Math.min(solutionRatio-0.1, 0);
        }

        if (solutionRatio > problemRatio) {
            return 0.5 + 0.5*Math.abs(solutionRatio - problemRatio);
        }
        else{
            return 0.5 - 0.5*Math.abs(solutionRatio - problemRatio);
        }
    }

    /*
        Terrain influences movement, melee and shooting phases.
        Difficult terrain impedes movement, on average the movement is impeded 70% of the time. Furthermore, Dangerous terrain is even more deadly, so moving through it is even more difficult.
        In shooting, terrain provides cover instead of armor. Thus this can further enhance or detrment units attempting to shoot.
        Finally, in melee charging through difficult terrain reduces the charge roll by 2, effectively bringing the average down from 2-12 to 0-10, which is about 16%.
     */
    public static double terrainRate(Army a, List<Double> terrain){

        /*
            First we determine (add or remove) the amount of buildings, difficult and dangerous terrain on the field.
            This is entirely optional for the players. The terrain can be fully open, or it can be almost fully enclosed with buildings.
            The majority of the terrain will (by default) is written in the terrain list, and this is taken as an average that can, and should, be changed.
         */
        double building = terrain.get(0)/100;
        double difficult = terrain.get(1)/100;
        double dangerous = terrain.get(2)/100;
        double impassable = terrain.get(3)/100;

        /*
            For movement most of the terrain limits/hampers movement, or simply nullifies it completely.
            We on average take the degree of hampering as a disadvantage, and see which unit is more hampered by terrain.
            Flyers, Skimmers, Jetbikes, Jump Units, Jetpack Units, and units with Move Through Cover rule ignore this penalty (mostly).
         */

        double hamper = 0;
        double solutionCount = 0;

        for (int i = 0; i < a.getSquad().size(); i++) {
            for (int j = 0; j < a.getSquad().get(i).getUnit().size(); j++) {
                if (a.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    solutionCount++;
                    for (int k = 0; k < a.getSquad().get(i).getUnit().get(j).getRules().size(); k++) {
                        if (a.getSquad().get(i).getUnit().get(j).getRules().get(k).equals("Move Through Cover")){
                            hamper += 1.0 / 6.0 * building + 1.0 / 6.0 * difficult + impassable;
                        }
                        else{
                            if (a.getSquad().get(i).getUnit().get(j).getType().contains("Skimmer") || a.getSquad().get(i).getUnit().get(j).getType().contains("Jetbike")
                                    || a.getSquad().get(i).getUnit().get(j).getType().contains("Jump") || a.getSquad().get(i).getUnit().get(j).getType().contains("Jetpack")) {
                                hamper += 1.0 / 6.0 * dangerous + 1.0 / 6.0 * impassable;
                            } else if (a.getSquad().get(i).getUnit().get(j).getType().contains("Flyer")) {
                                hamper += 0;
                            } else if (a.getSquad().get(i).getUnit().get(j).getType().equals("Infantry-Character") || a.getSquad().get(i).getUnit().get(j).getType().equals("Infantry") ||
                                    a.getSquad().get(i).getUnit().get(j).getType().equals("Artillery") || a.getSquad().get(i).getUnit().get(j).getType().contains("Walker")) {
                                hamper += 2.0 / 6.0 * building + 2.0 / 6.0 * difficult + 1.0 / 6.0 * dangerous + impassable;
                            } else if (a.getSquad().get(i).getUnit().get(j).getType().equals("Bike-Character") || a.getSquad().get(i).getUnit().get(j).getType().equals("Bike")) {
                                hamper += 1.0 / 6.0 * dangerous + impassable;
                            } else {
                                hamper += 1.0 / 6.0 * dangerous + impassable;
                            }
                        }
                    }
                }
            }
        }

        hamper = 1 - (hamper / solutionCount);

        return hamper;
    }

    /*
        This method calculates the average toughness, armour save and weapon skill to use in both assault and shooting phases.
     */
    public static double[] armyCalculation(Army a){

        double[] calculations = {0,0,0,0};
        int infC = 0;
        int vehC = 0;

        for (int i = 0; i < a.getSquad().size(); i++) {
            for (int j = 0; j < a.getSquad().get(i).getUnit().size(); j++) {
                if (a.getSquad().get(i).getUnit().get(j).getCount() > 0) {
                    if (a.getSquad().get(i).getUnit().get(j).getType().equals("Infantry") || a.getSquad().get(i).getUnit().get(j).getType().equals("Infantry-Character")
                            || a.getSquad().get(i).getUnit().get(j).getType().equals("Bike") || a.getSquad().get(i).getUnit().get(j).getType().equals("Bike-Character")
                            || a.getSquad().get(i).getUnit().get(j).getType().equals("Artillery")) {
                        infC+= a.getSquad().get(i).getUnit().get(j).getCount();
                        calculations[3] += a.getSquad().get(i).getUnit().get(j).getWS()*a.getSquad().get(i).getUnit().get(j).getCount();
                        calculations[2] += a.getSquad().get(i).getUnit().get(j).getAS()*a.getSquad().get(i).getUnit().get(j).getCount();
                        calculations[1] += a.getSquad().get(i).getUnit().get(j).getT()*a.getSquad().get(i).getUnit().get(j).getCount();
                        calculations[0] += a.getSquad().get(i).getUnit().get(j).getT()*a.getSquad().get(i).getUnit().get(j).getCount();
                    }
                    else {
                        vehC+= a.getSquad().get(i).getUnit().get(j).getCount();

                        if (a.getSquad().get(i).getUnit().get(j).getWS() == 0){
                            calculations[3]+=a.getSquad().get(i).getUnit().get(j).getCount();
                        }
                        else{
                            calculations[3] += a.getSquad().get(i).getUnit().get(j).getWS()*a.getSquad().get(i).getUnit().get(j).getCount();
                        }
                        calculations[1] += a.getSquad().get(i).getUnit().get(j).getRear()*a.getSquad().get(i).getUnit().get(j).getCount();
                        calculations[0] += ((a.getSquad().get(i).getUnit().get(j).getSide()+a.getSquad().get(i).getUnit().get(j).getSide())/2)*a.getSquad().get(i).getUnit().get(j).getCount();
                    }
                }
            }
        }

        if (infC > 0 || vehC > 0) calculations[0] = calculations[0] / (infC+vehC);
        if (infC > 0 || vehC > 0) calculations[1] = calculations[1] / (infC+vehC);
        if (infC > 0) calculations[2] = calculations[2] / infC;
        if (infC > 0 || vehC > 0) calculations[3] = calculations[3] / (infC+vehC);

        return calculations;
    }

    /*
        This method calculates the efficiency of an army (the number of wounds it inflicts) in the shooting phase.
     */
    public static double[] shootingCalculation(Army a, double avgToughness, double avgAS){

        double[] efficiency = {0,0};
        double ballistics;
        double hitChance;

        for (int i = 0; i < a.getSquad().size(); i++) {
            for (int j = 0; j < a.getSquad().get(i).getUnit().size(); j++) {
                if (a.getSquad().get(i).getUnit().get(j).getCount() > 0) {

                    ballistics = a.getSquad().get(i).getUnit().get(j).getBS();
                    if (ballistics == 0 && a.getSquad().get(i).getUnit().get(j).getType().equals("Artillery")){
                        ballistics = a.getSquad().get(i).getUnit().get(j+1).getBS();
                    }
                    List<Double> effectiveness = new ArrayList<>();
                    List<String> weaponNames = new ArrayList<>();
                    int unitC = (int) a.getSquad().get(i).getUnit().get(j).getCount();

                    if (ballistics < 6) {
                        hitChance = ballistics / 6.0;
                    } else {
                        hitChance = (5.0 / 6.0) + ((ballistics - 5.0) / 36.0);
                    }

                    for (int k = 0; k < a.getSquad().get(i).getUnit().get(j).getEquipment().size(); k++) {
                        for (int x = 0; x < CBRMain.EquipmentList.size(); x++) {
                            if (a.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getName().equals(CBRMain.EquipmentList.get(x).getName())) {
                                if (CBRMain.EquipmentList.get(x).getRange() > 0 || CBRMain.EquipmentList.get(x).getRange() == -1) {

                                    double r = CBRMain.EquipmentList.get(x).getRange();
                                    int s = Integer.parseInt(CBRMain.EquipmentList.get(x).getStrength());
                                    double armourp = CBRMain.EquipmentList.get(x).getAp();
                                    double effect;
                                    double armourEffect;
                                    double rangeEffect;
                                    double strengtheffect;

                                    if (r < 12) rangeEffect = 0.25;
                                    else if (r >= 12 && r < 24) rangeEffect = 0.5;
                                    else if (r >= 24 && r < 36) rangeEffect = 1.0;
                                    else rangeEffect = 2.0;

                                    if (CBRMain.EquipmentList.get(x).getType().contains("Twin-linked")){
                                        hitChance = hitChance + ((1-hitChance)*(hitChance));
                                    }

                                    if (r == -1){
                                        hitChance = 1.0;
                                    }

                                    //This weapon can not wound (and furthermore can not even dent armour), so the effectiveness is 0.
                                    if (s - avgToughness < -2) {
                                        strengtheffect = 0;
                                    }
                                    //This type of weapon is much stronger than toughness.
                                    else if (s - avgToughness >= 2) {
                                        strengtheffect = 5.0 / 6.0;
                                    }
                                    //This type of weapon is in between the above two values.
                                    else {
                                        strengtheffect = (s - avgToughness) * 0.1665 + 0.5;
                                    }

                                    if (armourp <= avgAS && armourp != 0) {
                                        armourEffect = 1.0;
                                    }
                                    else if (avgAS == 0){
                                        armourEffect = 1.0;
                                    }
                                    else {
                                        armourEffect = 1 - ((7.0 - avgAS) / 6.0);
                                    }
                                    effect = hitChance * strengtheffect * armourEffect * rangeEffect;
                                    effectiveness.add(effect * GeneralKnowledge.hitCount(CBRMain.EquipmentList.get(x).getType()));
                                    weaponNames.add(CBRMain.EquipmentList.get(x).getName());
                                }
                            }
                        }
                    }

                    if (!effectiveness.isEmpty()) {
                        int sort = 0;
                        while (sort < unitC) {
                            double best = -0.01;
                            int tmp = -1;
                            for (int k = 0; k < effectiveness.size(); k++) {
                                if (effectiveness.get(k) > best) {
                                    best = effectiveness.get(k);
                                    tmp = k;
                                }
                            }
                            for (int k = 0; k < a.getSquad().get(i).getUnit().get(j).getEquipment().size(); k++) {
                                if (weaponNames.get(tmp).equals(a.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getName())) {
                                    efficiency[0] += effectiveness.get(tmp) * a.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getCount();
                                    sort += a.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getCount();
                                    for (int x = 0; x < CBRMain.EquipmentList.size(); x++){
                                        if (weaponNames.get(tmp).equals(CBRMain.EquipmentList.get(x).getName()) && CBRMain.EquipmentList.get(x).getType().contains("Barrage")) {
                                            efficiency[1]+= effectiveness.get(tmp) * a.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getCount();
                                        }
                                    }
                                    effectiveness.remove(tmp);
                                }
                            }
                            if (effectiveness.isEmpty()) {
                                sort += unitC;
                            }
                        }
                    }
                }
            }
        }

        return efficiency;
    }

    /*
        This method calculates the efficiency of an army (the number of wounds it inflicts) in the shooting phase.
    */
    public static double assaultCalculation(Army a, double avgWS, double avgToughness, double avgAS){

        double efficiency = 0;
        double attack;
        double weaponskill;
        double strength;
        double hitChance = 0;

        for (int i = 0; i < a.getSquad().size(); i++) {
            for (int j = 0; j < a.getSquad().get(i).getUnit().size(); j++) {
                if (a.getSquad().get(i).getUnit().get(j).getCount() > 0) {

                    weaponskill = a.getSquad().get(i).getUnit().get(j).getWS();
                    strength = a.getSquad().get(i).getUnit().get(j).getS();
                    List<Double> effectiveness = new ArrayList<>();
                    List<String> weaponNames = new ArrayList<>();
                    int unitC = (int) a.getSquad().get(i).getUnit().get(j).getCount();

                    //calculate to hit
                    if (weaponskill > avgWS){
                        hitChance = 4.0/6.0;
                    }
                    else if ((weaponskill == avgWS) || (weaponskill < avgWS && weaponskill >= 5) || (weaponskill == 1 && avgWS < 2)
                            || (weaponskill == 2 && avgWS < 5) || (weaponskill == 3 && avgWS < 7)
                            || (weaponskill == 4 && avgWS < 9)){
                        hitChance = 3.0/6.0;
                    }
                    else if ((weaponskill == 1 && avgWS > 2) || (weaponskill == 2 && avgWS > 5)
                            || (weaponskill == 3 && avgWS > 7) || (weaponskill == 4 && avgWS > 9)){
                        hitChance = 2.0/6.0;
                    }

                    //Calculate effectiveness for weapons
                    for (int k = 0; k < a.getSquad().get(i).getUnit().get(j).getEquipment().size(); k++) {
                        for (int x = 0; x < CBRMain.EquipmentList.size(); x++) {
                            if (a.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getName().equals(CBRMain.EquipmentList.get(x).getName())) {
                                if (CBRMain.EquipmentList.get(x).getRange() == 0) {

                                    double effect;
                                    String s = CBRMain.EquipmentList.get(x).getStrength();
                                    double armourp = CBRMain.EquipmentList.get(x).getAp();
                                    double armourEffect;
                                    double strengthEffect = 0;
                                    double initiativeeffect = 0.1 * a.getSquad().get(i).getUnit().get(j).getI();

                                    if (initiativeeffect == 0){
                                        initiativeeffect = 0.1;
                                    }

                                    if (s.equals("+1")) {
                                        strength++;
                                    }
                                    if (s.equals("+2")) {
                                        strength += 2;
                                    }
                                    if (s.equals("+6")) {
                                        strength += 6;
                                    }
                                    if (s.equals("x2")) {
                                        strength *= 2;
                                    }
                                    if (strength > 10){
                                        strength = 10;
                                    }

                                    //This strength difference will not cause wounds.
                                    if (strength < 7 && avgToughness > strength + 4) {
                                        strengthEffect = 0;
                                    } else if (avgToughness - strength > 2) {
                                        strengthEffect = 1.0 / 6.0;
                                    } else if (avgToughness - strength < 2 && avgToughness - strength > 0) {
                                        strengthEffect = 2.0 / 6.0;
                                    } else if (avgToughness == strength) {
                                        strengthEffect = 3.0 / 6.0;
                                    } else if (strength - avgToughness < 2 && strength - avgToughness > 0) {
                                        strengthEffect = 4.0 / 6.0;
                                    } else if (strength - avgToughness >= 2) {
                                        strengthEffect = 5.0 / 6.0;
                                    }

                                    if (armourp <= avgAS && armourp != 0) {
                                        armourEffect = 1.0;
                                    }
                                    else if (avgAS == 0){
                                        armourEffect = 1.0;
                                    }
                                    else{
                                        armourEffect = 1 - ((7.0 - avgAS) / 6.0);
                                    }

                                    for (int z = 0; z < CBRMain.EquipmentList.get(x).getType().size(); z++){
                                        if (CBRMain.EquipmentList.get(x).getType().get(z).equals("Unwieldy")){
                                            initiativeeffect = 0.1;
                                        }
                                    }

                                    effect = hitChance * strengthEffect * armourEffect * initiativeeffect;
                                    effectiveness.add(effect * a.getSquad().get(i).getUnit().get(j).getA());
                                    weaponNames.add(CBRMain.EquipmentList.get(x).getName());
                                }
                            }
                        }
                    }

                    /*
                        Check to see if the list of weapons is empty. If the list is empty, the unit has no melee weapons
                        In this case the unit is still armed with a default melee weapon (not mentioned or included in the unit).
                     */

                    if (effectiveness.isEmpty()){

                        double effect;
                        double armourp = 0;
                        double armourEffect;
                        double strengthEffect = 0;
                        double initiativeeffect = 0.1 * a.getSquad().get(i).getUnit().get(j).getI();
                        attack = a.getSquad().get(i).getUnit().get(j).getA();

                        if (attack == 0){
                            attack = 1;
                        }

                        if (initiativeeffect == 0){
                            initiativeeffect = 0.1;
                        }

                        if (strength < 7 && avgToughness > strength + 4) {
                            strengthEffect = 0;
                        } else if (avgToughness - strength > 2) {
                            strengthEffect = 1.0 / 6.0;
                        } else if (avgToughness - strength < 2 && avgToughness - strength > 0) {
                            strengthEffect = 2.0 / 6.0;
                        } else if (avgToughness == strength) {
                            strengthEffect = 3.0 / 6.0;
                        } else if (strength - avgToughness < 2 && strength - avgToughness > 0) {
                            strengthEffect = 4.0 / 6.0;
                        } else if (strength - avgToughness >= 2) {
                            strengthEffect = 5.0 / 6.0;
                        }

                        if (armourp <= avgAS && armourp != 0) {
                            armourEffect = 1.0;
                        }
                        else if (avgAS == 0){
                            armourEffect = 1.0;
                        }
                        else{
                            armourEffect = 1 - ((7.0 - avgAS) / 6.0);
                        }

                        effect = hitChance * strengthEffect * armourEffect * initiativeeffect;
                        efficiency += effect * a.getSquad().get(i).getUnit().get(j).getCount() * attack;
                    }

                    if (!effectiveness.isEmpty() && !weaponNames.isEmpty()) {
                        int sort = 0;
                        while (sort < unitC) {
                            double best = -0.01;
                            int tmp = -1;
                            for (int k = 0; k < effectiveness.size(); k++) {
                                if (effectiveness.get(k) > best) {
                                    best = effectiveness.get(k);
                                    tmp = k;
                                }
                            }
                            for (int k = 0; k < a.getSquad().get(i).getUnit().get(j).getEquipment().size(); k++) {
                                if (weaponNames.get(tmp).equals(a.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getName())) {
                                    efficiency += effectiveness.get(tmp) * a.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getCount();
                                    sort += a.getSquad().get(i).getUnit().get(j).getEquipment().get(k).getCount();
                                    effectiveness.remove(tmp);
                                }
                            }
                            if (effectiveness.isEmpty()) {
                                sort += unitC;
                            }
                        }
                    }
                }
            }
        }

        return efficiency;
    }

    public static double shootingWeight(List<Double> terrain){

        double building = terrain.get(0)/100;
        double difficult = terrain.get(1)/100;
        double dangerous = terrain.get(2)/100;
        double impassable = terrain.get(3)/100;

          /*
            The shooting phase is rather simple.
            We assume that players will take absolute use of cover, so the cover bonus is counted as terrain * 1/3, as the terrain typically has cover 5+.
            This represents how much the units are hampered by shooting. This means that the shooting phase carries less weight.
         */

        return 1 - ((building + difficult + dangerous + impassable)*1.0/3.0);
    }

    public static double assaultWeight(List<Double> terrain){

        double difficult = terrain.get(1)/100;
        double dangerous = terrain.get(2)/100;

          /*
            In the assault phase, the more dangerous terrain the less likely charges are going to succeed.
            The effectiveness of the assault phase is hampered by it. The dice rolls for charge become 2d6-2, which is effectively 1/6 worse.
         */

        return 1 - ((difficult + dangerous)*1.0/6.0);
    }

    public static double ratioOfRating(Army a, Army p){
        return Math.pow(10, (a.getRating() / 400.0)) / (Math.pow(10, (a.getRating() / 400.0)) + Math.pow(10, (p.getRating() / 400.0)));
    }

    public static double ratioOfSquads(Army a, Army p){
        double aSquads = 0;
        for (int i = 0; i < a.getSquad().size(); i++){
            aSquads+=a.getSquad().get(i).getRating();
        }
        aSquads = aSquads / a.getSquad().size();

        double pSquads = 0;
        for (int i = 0; i < p.getSquad().size(); i++){
            pSquads+=p.getSquad().get(i).getRating();
        }
        pSquads = pSquads / p.getSquad().size();

        return Math.pow(10, (aSquads / 400.0)) / (Math.pow(10, (aSquads / 400.0)) + Math.pow(10, (pSquads / 400.0)));
    }

    /*
        This method handles the in-depth explanations.
     */
    public static void  explanations(Army a, Army p, double similarity, double bestARating, double  bestSRating, double  bestMoveRating, double  bestShootRating, double  bestPsyRating,
                                     double bestAssaultRating, double  bestFavourRating){

        CBRMain.retrievalExplanations.add("In depth retrieval explanation:");
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("The win chance is calculated based on the following formula: (Army Rating + Squad Rating + Phases Rating*6 + Favour Rating*3)/11");
        CBRMain.retrievalExplanations.add("Keep in mind that percentile values presented here are not modified with the contribution (weight) to the total win chance.");
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("Army Rating:");
        CBRMain.retrievalExplanations.add("Army rating contributes to: " + String.format("%.4g", (1.0/11.00)*100.0) + "% of the total win chance.");
        CBRMain.retrievalExplanations.add("");
        if (bestARating > 0.5) {
            CBRMain.retrievalExplanations.add("The retrieved army has a rating of: " + a.getRating() + "." + " The problem army has a rating of "
                    + p.getRating() + ".\n" + "The solution army therefore has a " + String.format("%.4g", (bestARating-0.5)*100.0) + "% advantage due to rating.");
        }
        else if(bestARating == 0.5){
            CBRMain.retrievalExplanations.add("The retrieved army has a rating of: " + a.getRating() + "." + " The problem army has a rating of "
                    + p.getRating() + ".\n" + "Neither of the armies has an advantage due to Army Ratings.");

        }
        else{
            CBRMain.retrievalExplanations.add("The retrieved army has a rating of: " + a.getRating() + "." + " The problem army has a rating of "
                    + p.getRating() + ".\n" + "The solution army therefore has a " + String.format("%.4g", (0.5-bestARating)*100.0) + "% disadvantage due to rating.");
        }
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("Squad Rating:");
        CBRMain.retrievalExplanations.add("Squad rating contributes to: " + String.format("%.4g", (1.0/11.00)*100.0) + "% of the total win chance.");
        CBRMain.retrievalExplanations.add("");
        if (bestSRating > 0.5) {
            CBRMain.retrievalExplanations.add("When comparing the retrieved army with the problem army based on average squad rating, the solution army has a "
                    + String.format("%.4g", (bestSRating-0.5)*100.0) + "% advantage due to squad rating.");
        }
        else if(bestSRating == 0.5){
            CBRMain.retrievalExplanations.add("Neither of the armies has an advantage due to squad rating.");

        }
        else{
            CBRMain.retrievalExplanations.add("When comparing the retrieved army with the problem army based on average squad rating, the solution army has a "
                    + String.format("%.4g", (0.5-bestSRating)*100.0) + "% disadvantage due to squad rating.");
        }
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("Phases Rating:");
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("The ratings for each phase of combat were calculated. The Shooting and Movement phases have a greater weight in this.");
        CBRMain.retrievalExplanations.add("The exact formula is (Movement*" + CBRMain.movementWeight + " + Psychic*" + CBRMain.psychicWeight +" + Shooting*" + CBRMain.shootingWeight + " + Assault*"
                                            + CBRMain.assaultWeight +")/" + (CBRMain.movementWeight+CBRMain.psychicWeight+CBRMain.shootingWeight+CBRMain.assaultWeight) + ".");
        CBRMain.retrievalExplanations.add("This calculation is considered the most important and contributes to: " + String.format("%.4g", (6.0/11.0)*100.0) + "% of the total win chance.");
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("Movement:");
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("For movement, the calculation is based on the combined rate of the total army movement.");
        CBRMain.retrievalExplanations.add("Of that, 90% of movement is considered movement in the Movement Phase, and 10% is considered the Run action.");
        CBRMain.retrievalExplanations.add("The terrain is included in the calculation. Units that avoid hazardous terrains have an advantage as the rules state.");
        CBRMain.retrievalExplanations.add("Transports attached to infantry units are assumed to transport units. " +
                                 "\nTransports not attached to any infantry unit are assumed to still perform their transport function.");
        if (bestMoveRating > 0.5) {
            CBRMain.retrievalExplanations.add("Based on the calculations, the solution army has a " + String.format("%.4g", 1.1*(bestMoveRating-0.5)*100.0) + "% advantage due to combined movement ratings.");
        }
        else if(bestSRating == 0.5){
            CBRMain.retrievalExplanations.add("Neither of the armies has an advantage due to combined movement ratings.");

        }
        else{
            CBRMain.retrievalExplanations.add("Based on the calculations, the solution army has a " + String.format("%.4g", 1.1*(bestMoveRating-0.5)*100.0) + "% disadvantage due to combined movement ratings.");
        }

        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("Psychic:");
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("In the Psychic phase, the total amount of units with the Psyker, Psychic Pilot, Brotherhood of Psykers and Brotherhood of Sorcerers special abilities.");
        CBRMain.retrievalExplanations.add("We then directly compare the strengths of the psykers in each army to obtain the psychic strength rating.");
        if (bestPsyRating > 0.5) {
            CBRMain.retrievalExplanations.add("Based on the calculations, the solution army has a " + String.format("%.4g", 0.9*(bestPsyRating-0.5)*100.0) + "% advantage in the psychic phase.");
        }
        else if(bestPsyRating == 0.5){
            CBRMain.retrievalExplanations.add("Neither of the armies has an advantage due to psychic ratings.");

        }
        else{
            CBRMain.retrievalExplanations.add("Based on the calculations, the solution army has a " + String.format("%.4g", 0.9*(0.5-bestPsyRating)*100.0) + "% disadvantage in the psychic phase.");
        }
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("Shooting:");
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("In the shooting phase, we calculate the average of armour, toughness and armour saves for both armies.");
        CBRMain.retrievalExplanations.add("Then, each unit is checked for its ballistics skill and all of its available weapons.");
        CBRMain.retrievalExplanations.add("Each squad gets an effectiveness rating based on the combined effectiveness against the opposing army of each unit in that squad.");
        CBRMain.retrievalExplanations.add("The special abilities of weapons that are currently counted are: Twin-Linked, Blast, Large Blast and Rapid Fire. The number of shots a weapon has is also counted.");
        if (bestShootRating > 0.5) {
            CBRMain.retrievalExplanations.add("Based on the calculations, the solution army has a " + String.format("%.4g", 1.1*(bestShootRating-0.5)*100.0) + "% advantage in the shooting phase.");
        }
        else if(bestShootRating  == 0.5){
            CBRMain.retrievalExplanations.add("Neither of the armies has an advantage due to their strengths in the shooting phase.");

        }
        else{
            CBRMain.retrievalExplanations.add("Based on the calculations, the solution army has a " + String.format("%.4g", 1.1*(0.5-bestShootRating)*100.0) + "% disadvantage in the shooting phase.");
        }
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("Assault:");
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("The assault phase works similarly to the shooting phase, in that the average of armour, toughness, weapon skill and armour saves for both armies are counted.");
        CBRMain.retrievalExplanations.add("Then, each unit is checked for its weapon skill versus the enemy army and all of its available melee weapons.");
        CBRMain.retrievalExplanations.add("Each squad gets an effectiveness rating based on the combined effectiveness against the opposing army of each unit in that squad.");
        CBRMain.retrievalExplanations.add("The special abilities of weapons that are currently counted are: Unwieldy. Other inherent abilities, such as increase in strength is also counted.");
        if (bestAssaultRating > 0.5) {
            CBRMain.retrievalExplanations.add("Based on the calculations, the solution army has a " + String.format("%.4g", 0.9*(bestAssaultRating-0.5)*100.0) + "% advantage in the assault phase.");
        }
        else if(bestAssaultRating  == 0.5){
            CBRMain.retrievalExplanations.add("Neither of the armies has an advantage due to their strengths in the assault phase.");

        }
        else{
            CBRMain.retrievalExplanations.add("Based on the calculations, the solution army has a " + String.format("%.4g", 0.9*(0.5-bestAssaultRating)*100.0) + "% disadvantage in the assault phase.");
        }

        double tempTotal = (bestMoveRating*CBRMain.movementWeight + bestPsyRating*CBRMain.psychicWeight + bestShootRating*CBRMain.shootingWeight + bestAssaultRating*CBRMain.assaultWeight)/(CBRMain.movementWeight+CBRMain.psychicWeight+CBRMain.shootingWeight+CBRMain.assaultWeight);
        CBRMain.retrievalExplanations.add("");
        if (tempTotal > 0.5) {
            CBRMain.retrievalExplanations.add("The combined rating of the four phases gives the solution army a " + String.format("%.4g", (tempTotal-0.5)*100.0) + "% advantage in upcoming battle.");
        }
        else if(tempTotal  == 0.5){
            CBRMain.retrievalExplanations.add("The combined rating of the four phases gives neither army an advantage in the upcoming battle.");

        }
        else{
            CBRMain.retrievalExplanations.add("The combined rating of the four phases gives the solution army a " + String.format("%.4g", (0.5-tempTotal)*100.0) + "% disadvantage in upcoming battle.");
        }
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("Only armies that are fully in detachments are classified as solution armies. This also counts the warlord unit.");
        CBRMain.retrievalExplanations.add("The problem army is checked to see if it is fully in a detachment and if it has a valid warlord.");
        CBRMain.retrievalExplanations.add("This rating contributes to: " + String.format("%.4g", (3.0/11.00)*100.0) + "% of the total win chance.");
        CBRMain.retrievalExplanations.add("");
        if (bestFavourRating > 0.5){
            CBRMain.retrievalExplanations.add("The solution army has an advantage, as the problem army has units outside of detachments or lacks a warlord. The advantage is: " + String.format("%.4g", bestFavourRating*100.0) + "%.");
        }
        else if(bestFavourRating  == 0.5){
            CBRMain.retrievalExplanations.add("Both armies are fully in detachments with warlords. Neither army has an advantage.");

        }
        CBRMain.retrievalExplanations.add("");
        CBRMain.retrievalExplanations.add("Based on these factors, the win chance for the solution army is: " + String.format("%.4g", similarity*100.0) + "%.");
        CBRMain.retrievalExplanations.add("Some round off errors may be present in the explanations, up to +-0.2%.");
        CBRMain.retrievalExplanations.add("");
    }
}
